from psycopg2.extras import DateTimeTZRange
from sqlalchemy.dialects.postgresql import TSTZRANGE
from sqlalchemy.sql.expression import case, select, text
from sqlalchemy.sql import func
import requests
import zipfile
from io import BytesIO
from xml.dom.minidom import parseString
from datetime import date, datetime, time, timedelta

from indice_pollution.extensions import db
from indice_pollution.history.models.departement import Departement
from indice_pollution.history.models.commune import Commune

class VigilanceMeteo(db.Model):

    id = db.Column(db.Integer, primary_key=True)

    zone_id = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'))
    phenomene_id = db.Column(db.Integer)
    date_export = db.Column(db.DateTime)

    couleur_id = db.Column(db.Integer)
    validity = db.Column(TSTZRANGE(), nullable=False)

    __table_args__ = (
        db.Index('vigilance_zone_phenomene_date_export_idx', zone_id, phenomene_id, date_export),
        db.UniqueConstraint(zone_id, phenomene_id, date_export, validity),
        {"schema": "indice_schema"},
    )

    couleurs = {
        1: 'Vert',
        2: 'Jaune',
        3: 'Orange',
        4: 'Rouge'
    }

    labels = {
        1: 'Vigilance verte',
        2: 'Vigilance jaune',
        3: 'Vigilance orange',
        4: 'Vigilance rouge'
    }

    phenomenes = {
        1: 'Vent violent',
        2: 'Pluie-Inondation',
        3: 'Orages',
        4: 'Crues',
        5: 'Neige-verglas',
        6: 'Canicule',
        7: 'Grand Froid',
        8: 'Avalanches',
        9: 'Vagues-Submersion'
    }

    hours_to_add_before_6 = 16
    hours_to_add_before_16 = 30
    hours_to_add_after_16 = 40

    @staticmethod
    def get_departement_codes(code):
        if code == "20":
            return ["2A"]
        elif code == "120":
            return ["2B"]
        elif code == "175":
            return ["75", "92", "93", "94"]
        elif code == "99":
            return [None]
        return [code]

    @classmethod
    def save_all(cls):
        def convert_datetime(a):
            return datetime.strptime(a.value, "%Y%m%d%H%M%S")
        r = requests.get("http://vigilance2019.meteofrance.com/data/vigilance.zip")
        with zipfile.ZipFile(BytesIO(r.content)) as z:
            fname = "NXFR49_LFPW_.xml"
            if not fname in z.namelist():
                return
            with z.open(fname) as f:
                x = parseString(f.read())
                date_export = convert_datetime(x.getElementsByTagName('SIV_MENHIR')[0].attributes['dateExportTU'])
                if db.session.query(func.max(cls.date_export)).first() == (date_export,):
                    return

                for phenomene in x.getElementsByTagName("PHENOMENE"):
                    for departement_code in cls.get_departement_codes(phenomene.attributes['departement'].value):
                        if not departement_code:
                            continue
                        departement = Departement.get(departement_code)
                        if not departement:
                            continue
                        debut = convert_datetime(phenomene.attributes['dateDebutEvtTU'])
                        fin = convert_datetime(phenomene.attributes['dateFinEvtTU'])
                        obj = cls(
                            zone_id=departement.zone_id,
                            phenomene_id=int(phenomene.attributes['phenomene'].value),
                            date_export=date_export,
                            couleur_id=int(phenomene.attributes['couleur'].value),
                            validity=DateTimeTZRange(debut, fin),
                        )
                        db.session.add(obj)
                db.session.commit()

    # Cette requête selectionne les vigilances météo du dernier export fait avant date_ & time_
    # Et ne renvoie que les vigilances comprises entre :
    #  * le date & time passés et date_export J +16h si l’heure dans le date export est < 6
    #  * le date & time passés et date_export J+1 6h si l’heure dans le date export est < 16
    #  * le date & time passés et date_export J+1 16h si l’heure dans le date export est >= 16
    @classmethod
    def get_query(cls, departement_code, insee, date_, time_):
        if not isinstance(date_, date):
            datetime_ = datetime.now()
        else:
            datetime_ = datetime.combine(date_, datetime.now().time())
        if isinstance(time_, time):
            datetime_ = datetime.combine(date_, time_)

        # On est obligé de passer par les tables sinon la fonction fetch à la fin n’est pas prise en compte
        vigilance_t = VigilanceMeteo.__table__
        departement_t = Departement.__table__
        statement = select(
            vigilance_t
        ).join(
            departement_t, vigilance_t.c.zone_id == departement_t.c.zone_id
        ).filter(
            vigilance_t.c.date_export <= datetime_,
            vigilance_t.c.validity.overlaps(
                DateTimeTZRange(
                    cls.make_start_date(date_=datetime_),
                    cls.make_end_date(date_=datetime_)
                )
            ),
            # Permet de ne pas selectionner des vigilances publiées il y a trop longtemps
            # ça donne des données parcelaires dont on ne veut pas.
            func.date_trunc('day', vigilance_t.c.date_export) + case(
                ((func.date_part('hour', vigilance_t.c.date_export) < 6), text(f"'{cls.hours_to_add_before_6}h'::interval")),
                ((func.date_part('hour', vigilance_t.c.date_export) < 16), text(f"'{cls.hours_to_add_before_16}h'::interval")),
                else_=text(f"'{cls.hours_to_add_after_16}h'::interval")
            ) > datetime_
        ).order_by(
            vigilance_t.c.date_export.desc()
        )

        if insee:
            commune_t = Commune.__table__
            statement = statement.join(
                commune_t, departement_t.c.id == commune_t.c.departement_id
            ).filter(commune_t.c.insee == insee)
        elif departement_code:
            statement = statement.filter(departement_t.c.code == departement_code)

        return statement.fetch(1, with_ties=True) # Renvoie toutes les vigilances avec le même date export, voir FECTH LAST 1 ROW WITH TIES

    @classmethod
    def get(cls, departement_code=None, insee=None, date_=None, time_=None):
        orms_obj = select(cls).from_statement(cls.get_query(departement_code, insee, date_, time_))
        return list(db.session.execute(orms_obj).scalars())

    @classmethod
    def get_all(cls, date_=None, time_=None):
        orms_obj = select(cls).from_statement(cls.get_query(None, None, date_, time_))
        return list(db.session.execute(orms_obj).scalars())

    @property
    def couleur(self) -> str:
        return self.couleurs.get(self.couleur_id)

    @property
    def phenomene(self) -> str:
        return self.phenomenes.get(self.phenomene_id)

    def __repr__(self) -> str:
        return f"<VigilanceMeteo zone_id={self.zone_id} phenomene_id={self.phenomene_id} date_export={self.date_export} couleur_id={self.couleur_id} validity={self.validity}>"

    @classmethod
    def make_max_couleur(cls, vigilances):
        couleurs = [v.couleur_id for v in vigilances]
        return max(couleurs) if couleurs else None

    @classmethod
    def make_label(cls, max_couleur=None):
        return VigilanceMeteo.labels.get(max_couleur, '')

    @classmethod
    def make_start_date(cls, vigilances=None, date_=None):
        if isinstance(date_, datetime):
            return cls.make_end_date(date_=date_) - timedelta(days=1)
        if isinstance(vigilances, list) and len(vigilances) > 0:
            return vigilances[0].date_export
        return None

    @classmethod
    def make_end_date(cls, vigilances=None, date_=None):
        if date_ is None:
            if not isinstance(vigilances, list) or len(vigilances) < 1:
                return None
            date_ = vigilances[0].date_export
        if not isinstance(date_, datetime):
            return None

        if date_.hour < 6:
            hours_to_add = cls.hours_to_add_before_6
        elif date_.hour < 16:
            hours_to_add = cls.hours_to_add_before_16
        else:
            hours_to_add = cls.hours_to_add_after_16

        return date_.replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(hours=hours_to_add)



from celery import Celery
from flask_migrate import Migrate
from flask_caching import Cache

try:
    from flask_sqlalchemy import SQLAlchemy
except ImportError:
    db = None
else:
    db = SQLAlchemy()

celery = Celery(__name__)
cache = Cache()
migrate = Migrate()


from itertools import groupby
from sqlalchemy.orm import joinedload
from indice_pollution.history.models.commune import Commune
from indice_pollution.history.models.indice_atmo import IndiceATMO
from indice_pollution.history.models.episode_pollution import EpisodePollution
from indice_pollution.history.models.indice_uv import IndiceUv
from flask import Flask
from flask_cors import CORS
from datetime import date
import os
import logging
from indice_pollution.history.models.raep import RAEP
from indice_pollution.history.models.vigilance_meteo import VigilanceMeteo

from .helpers import today
from .extensions import celery, cache, db, migrate
from importlib import import_module
from kombu import Queue
from celery.schedules import crontab

def configure_celery(flask_app):
    """Configure tasks.celery:
    * read configuration from flask_app.config and update celery config
    * create a task context so tasks can access flask.current_app
    Doing so is recommended by flask documentation:
    https://flask.palletsprojects.com/en/1.1.x/patterns/celery/
    """
    # Settings list:
    # https://docs.celeryproject.org/en/stable/userguide/configuration.html
    celery_conf = {
        key[len('CELERY_'):].lower(): value
        for key, value in flask_app.config.items()
        if key.startswith('CELERY_')
    }
    celery.conf.update(celery_conf)
    celery.conf.task_queues = (
        Queue("default", routing_key='task.#'),
        Queue("save_indices", routing_key='save_indices.#'),
    )
    celery.conf.task_default_exchange = 'tasks'
    celery.conf.task_default_exchange_type = 'topic'
    celery.conf.task_default_routing_key = 'task.default'

    class ContextTask(celery.Task):
        def __call__(self, *args, **kwargs):
            with flask_app.app_context():
                return self.run(*args, **kwargs)

    celery.Task = ContextTask

def configure_cache(app):
    conf = {
        "CACHE_TYPE": os.getenv("CACHE_TYPE", "SimpleCache"),
        "CACHE_DEFAULT_TIMEOUT": os.getenv("CACHE_DEFAULT_TIMEOUT", 86400),
    }
    if conf["CACHE_TYPE"] == "RedisCache":
        conf["CACHE_REDIS_HOST"] = os.getenv("REDIS_HOST")
        conf["CACHE_REDIS_PASSWORD"] = os.getenv("REDIS_PASSWORD")
        conf["CACHE_REDIS_PORT"] = os.getenv("REDIS_PORT")
        conf["CACHE_REDIS_TOKEN"] = os.getenv("REDIS_TOKEN")
        conf["CACHE_REDIS_URL"] = os.getenv("REDIS_URL")
    elif conf["CACHE_TYPE"] == "SimpleCache":
        conf["CACHE_THRESHOLD"] = 100000
    cache.init_app(app, conf)

@celery.task()
def save_all_indices():
    save_all()

@celery.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    sender.add_periodic_task(
        crontab(minute='0', hour='*/1'),
        save_all_indices.s(),
        queue='save_indices',
        routing_key='save_indices.save_all'
    )

def init_app(app):
    db.init_app(app)
    configure_celery(app)
    configure_cache(app)

def create_app(test_config=None):
    app = Flask(
        __name__,
        instance_relative_config=True,
        static_url_path=''
    )
    app.config.from_mapping(
        SECRET_KEY=os.getenv('SECRET_KEY', 'dev'),
        SQLALCHEMY_DATABASE_URI=os.getenv('SQLALCHEMY_DATABASE_URI') or os.getenv('POSTGRESQL_ADDON_URI'),
        SQLALCHEMY_TRACK_MODIFICATIONS=False
    )
    app.config['CELERY_RESULT_BACKEND'] = os.getenv('CELERY_RESULT_BACKEND') or f"db+{app.config['SQLALCHEMY_DATABASE_URI']}"
    app.config['CELERY_BROKER_URL'] = os.getenv('CELERY_BROKER_URL') or f"sqla+{app.config['SQLALCHEMY_DATABASE_URI']}"

    CORS(app, send_wildcard=True)

    init_app(app)
    migrate.init_app(app, db)

    with app.app_context():
        import indice_pollution.api
        import indice_pollution.history

    return app

def make_resp(r, result, date_=None):
    if type(result) == list:
        if date_:
            result = [v for v in result if v['date'] == str(date_)]
    elif hasattr(result, 'dict'):
        result = [result.dict()]
    else:
        result = [result]
    return {
        "data": result,
        "metadata": make_metadata(r)
    }

def make_metadata(r):
    return {
        "region": {
            "nom": r.__name__.split(".")[-1],
            "website": r.Service.website,
            "nom_aasqa": r.Service.nom_aasqa
        }
    }

def forecast(insee, date_=None, use_make_resp=True):
    from .regions.solvers import get_region
    date_ = date_ or today()
    try:
        region = get_region(insee)
    except KeyError:
        return {
            "error": f"No region for {insee}",
            "metadata": {}
        }, 400
    if region.Service.is_active:
        indice = IndiceATMO.get(insee=insee, date_=date_)
        if use_make_resp:
            return make_resp(region, indice, date_)
        else:
            if indice is not None:
                indice.region = region
                indice.commune = Commune.get(insee)
            return indice
    else:
        indice = IndiceATMO()
        indice.region = region
        indice.commune = Commune.get(insee)
        indice.error = "Inactive region"
        return indice


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def get_all(date_):
    from indice_pollution.history.models import IndiceATMO, EpisodePollution
    date_ = date_ or today()
    indices = IndiceATMO.get_all(date_)
    episodes = EpisodePollution.get_all(date_)
    allergenes_par_departement = {
            r.zone_id: r
            for r in RAEP.get_all()
        }
    vigilances_par_departement = {
        zone_id: list(vigilances)
        for zone_id, vigilances in groupby(
            sorted(
                VigilanceMeteo.get_all(date_=date_),
                key=lambda v: v.zone_id
            ),
            lambda v: v.zone_id
        )
    }
    indices_uv = IndiceUv.get_all(date_)
    return (indices, episodes, allergenes_par_departement, vigilances_par_departement, indices_uv)

def bulk(insee_region_names: dict(), date_=None, fetch_episodes=False, fetch_allergenes=False):
    from indice_pollution.history.models import IndiceATMO, EpisodePollution
    from .regions.solvers import get_region
    date_ = date_ or today()

    insees = set(insee_region_names.keys())
    insees_errors = set()
    for insee in insees:
        try:
            region = get_region(region_name=insee_region_names[insee])
            if not region.Service.is_active:
                insees_errors.add(insee)
                continue
        except KeyError:
            insees_errors.add(insee)
            continue
    for insee in insees_errors:
        insees.remove(insee)
        del insee_region_names[insee]

    indices = dict()
    episodes = dict()
    for chunk in chunks(list(insees), 100):
        indices.update(
            {i['insee']: IndiceATMO.make_dict(i) for i in IndiceATMO.bulk(date_=date_, insees=chunk)}
        )
        if fetch_episodes:
            episodes.update(
                {e['insee']: dict(e) for e in EpisodePollution.bulk(date_=date_, insees=chunk)}
            )
    to_return = {
        insee: {
            "forecast": make_resp(
                get_region(region_name=insee_region_names[insee]),
                indices.get(insee, [])
               ),
            **({
                "episode": make_resp(
                    get_region(region_name=insee_region_names[insee]),
                    episodes.get(insee, [])
                )
                } if fetch_episodes else {}
            )
        }
        for insee in insees
    }
    if fetch_allergenes:
        allergenes_par_departement = {
            r.zone_id: r
            for r in RAEP.get_all()
        }
        communes = {
            c.insee: c
            for c in Commune.query.options(
                    joinedload(Commune.departement)
                ).populate_existing(
                ).all()
        }
        for insee in insees:
            c = communes[insee]
            if c.departement.zone_id in allergenes_par_departement:
                to_return.setdefault(insee, {})
                to_return[insee].update({'raep': allergenes_par_departement[c.departement.zone_id].to_dict()})
    return to_return

def episodes(insee, date_=None, use_make_resp=True):
    from .regions.solvers import get_region
    date_ = date_ or today()
    if type(date_) == str:
        date_ = date.fromisoformat(date_)

    region = get_region(insee)
    if region.Service.is_active:
        result = EpisodePollution.get(insee=insee, date_=date_)
        if use_make_resp:
            result = list(map(lambda e: e.dict(), result))
            return make_resp(region, result, date_)
        else:
            for r in result:
                r.region = region
                r.commune = Commune.get(insee)
            return result
    else:
        if use_make_resp:
            return {
                "error": "Inactive region",
                "metadata": make_metadata(region)
            }, 400
        else:
            episode = EpisodePollution()
            episode.region = region
            episode.commune = Commune.get(insee)
            episode.error = "Inactive region"
            return [episode]

def availability(insee):
    from .regions.solvers import get_region
    try:
        return get_region(insee).Service.is_active
    except KeyError:
        return False
    except AttributeError:
        return False


def raep(insee, date_=None, extended=False):
    if insee is None:
        return {}
    departement = Commune.get(insee).departement
    if not departement:
        return {}
    data = RAEP.get(zone_id=departement.zone_id, date_=date_)
    return {
        "departement": {
            "nom": departement.nom,
            "code": departement.code
        },
        "data": data.to_dict() if data else None
    }


def save_all():
    logger = logging.getLogger(__name__)
    logger.info('Begin of "save_all" task')
    regions = [
        'Auvergne-Rhône-Alpes',
        'Bourgogne-Franche-Comté',
        'Bretagne',
        'Centre-Val de Loire',
        'Corse',
        'Grand Est',
        'Guadeloupe',
        'Guyane',
        'Hauts-de-France',
        'Île-de-France',
        'Martinique',
        'Mayotte',
        'Normandie',
        'Nouvelle-Aquitaine',
        'Occitanie',
        'Pays de la Loire',
        "Réunion",
        "Sud"
    ]
    for region in regions:
        logger.info(f'Saving {region} region')
        module = import_module(f"indice_pollution.regions.{region}")
        if not module.Service.is_active:
            continue
        logger.info(f'Saving Forecast of {region}')
        module.Forecast().save_all()
        logger.info(f'Saving Episode of {region}')
        module.Episode().save_all()
        logger.info(f'Saving of {region} ended')
    RAEP.save_all()
    VigilanceMeteo.save_all()
    IndiceUv.save_all()


from indice_pollution.extensions import db
from indice_pollution.history.models.region import Region
from indice_pollution.history.models.tncc import TNCC
from sqlalchemy.orm import relation, relationship
import requests

class Departement(db.Model, TNCC):
    __table_args__ = {"schema": "indice_schema"}
    __tablename__ = 'departement'

    id = db.Column(db.Integer, primary_key=True)
    nom = db.Column(db.String)
    code = db.Column(db.String)
    region_id = db.Column(db.Integer, db.ForeignKey('indice_schema.region.id'))
    region = relationship("indice_pollution.history.models.region.Region")
    preposition = db.Column(db.String)
    zone_id = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'))
    zone = relationship("indice_pollution.history.models.zone.Zone")

    def __init__(self, **kwargs):
        if 'codeRegion' in kwargs:
            self.region = Region.get(kwargs.pop('codeRegion'))
        super().__init__(**kwargs)

    @classmethod
    def get(cls, code):
        return db.session.query(cls).filter_by(code=code).first() or cls.get_and_init_from_api(code)

    @classmethod
    def get_and_init_from_api(cls, code):
        res_api = cls.get_from_api(code)

        r = cls(**res_api)
        db.session.add(r)
        db.session.commit()
        return r

    @classmethod
    def get_from_api(cls, code):
        r = requests.get(
            f'https://geo.api.gouv.fr/departements/{code}?fields=nom,code,codeRegion',
            headers={"Accept": "application/json"}
        )
        r.raise_for_status()
        return r.json()


from indice_pollution.extensions import db

class TNCC:
    tncc = db.Column(db.Integer)
    tncc_codes= {
        0: {"article": " ",    "charniere": "de "},
        1: {"article": " ", 	  "charniere": "d'"},
        2: {"article": "le ",  "charniere": "du "},
        3: {"article": "la ",  "charniere": "de la "},
        4: {"article": "les ", "charniere": "des "},
        5: {"article": "l’",  "charniere": "de l’"},
        6: {"article": "aux ", "charniere": "des "},
        7: {"article": "las ", "charniere": "de las "},
        8: {"article": "los ", "charniere": "de los "}
    }

    def nom_charniere(self):
        return f"{self.tncc_codes[self.tncc]['charniere']}{self.nom}"


from indice_pollution.extensions import db
from indice_pollution.helpers import today
from indice_pollution.history.models.commune import Commune
from dataclasses import dataclass
from datetime import datetime
from sqlalchemy.dialects.postgresql import insert as pg_insert
from ftplib import FTP
import os, logging, csv, io

@dataclass
class IndiceUv(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    zone_id: int = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'), primary_key=True)
    date: datetime.date = db.Column(db.Date, primary_key=True, nullable=False)
    uv_j0: int = db.Column(db.Integer)
    uv_j1: int = db.Column(db.Integer)
    uv_j2: int = db.Column(db.Integer)
    uv_j3: int = db.Column(db.Integer)

    @classmethod
    def save_all(cls):

        def get_ftp_content():
            if not os.getenv('FS_BUCKET_URL') or not os.getenv('FS_BUCKET_USERNAME') or not os.getenv('FS_BUCKET_PASSWORD'):
                logging.error('The following env vars are required: FS_BUCKET_URL, FS_BUCKET_USERNAME and FS_BUCKET_PASSWORD')
                return None
            try:
                ftp = FTP(os.getenv('FS_BUCKET_URL'))
                ftp.login(os.getenv('FS_BUCKET_USERNAME'), os.getenv('FS_BUCKET_PASSWORD'))
                bio = io.BytesIO()
                filename = f'{today().strftime("%Y%m%d")}'
                ftp.retrbinary(f'RETR {filename}.csv', callback=bio.write)
                b = bio.getvalue()
                decoded_content = b.decode('UTF-8')
                return decoded_content
            except Exception as e:
                logging.error(e)
                return None

        decoded_content = get_ftp_content()
        if decoded_content is None:
            return
        reader = csv.DictReader(
            decoded_content.splitlines(),
            delimiter=';'
        )
        indices = []

        def format_insee(raw_insee: str):
            if len(raw_insee) == 5 and raw_insee.isdigit():
                return raw_insee
            else:
                return None
        def format_uv(raw_uv: str): # '-999' or '' means no value
            if raw_uv.isdigit():
                uv = int(raw_uv)
                if uv < 0:
                    uv = None
            else:
                uv = None
            return uv

        # file header: insee;commune;date;UV_J0;UV_J1;UV_J2;UV_J3
        for row in reader:
            insee = format_insee(row['insee'])
            outdated_communes = []
            if insee and insee not in outdated_communes:
                departement_code = f'{insee[:2]}'
                if departement_code == '20': # Corse: 20 in file, 2A or 2B in db
                    insee_2a = insee.replace('20', '2A', 1)
                    commune = Commune.get(insee_2a)
                    if commune is None:
                        insee_2b = insee.replace('20', '2B', 1)
                        commune = Commune.get(insee_2b)
                else:
                    commune = Commune.get(insee)
                date_format = '%Y-%m-%d'
                try:
                    date = datetime.strptime(row['date'], date_format).date()
                except Exception as e:
                    logging.error(e)
                    continue
                if commune and date:
                    indices.append({
                        **{
                            "zone_id": commune.zone_id,
                            "date": date,
                            "uv_j0": format_uv(row['UV_J0']),
                            "uv_j1": format_uv(row['UV_J1']),
                            "uv_j2": format_uv(row['UV_J2']),
                            "uv_j3": format_uv(row['UV_J3'])
                        }
                    })
        ins = pg_insert(cls)\
            .values(indices)\
            .on_conflict_do_nothing()
        db.session.execute(ins)
        db.session.commit()

    @classmethod
    def get(cls, insee, date_=None):
        date_ = date_ or today()
        query = cls.query.join(Commune, Commune.zone_id == cls.zone_id).filter(Commune.insee == insee, IndiceUv.date==date_)
        return query.first()

    @classmethod
    def get_all_query(cls, date_):
        return db.session.query(Commune.id, IndiceUv).join(Commune, Commune.zone_id == cls.zone_id).filter(IndiceUv.date==date_)

    @classmethod
    def get_all(cls, date_=None):
        date_ = date_ or today()
        return dict(cls.get_all_query(date_).all())

    @property
    def label(self):
        if type(self.uv_j0) == int:
            if self.uv_j0 >= 11:
                label = 'Extrême'
            elif self.uv_j0 >= 8:
                label = 'Très\u00a0fort' # unbreakable space
            elif self.uv_j0 >= 6:
                label = 'Fort'
            elif self.uv_j0 >= 3:
                label = 'Modéré'
            elif self.uv_j0 >= 1:
                label = 'Faible'
            else:
                label = 'Nul'
            label += f' (UV\u00a0{self.uv_j0})' # unbreakable space
        else:
            label = None
        return label


from indice_pollution import today
from indice_pollution.history.models import IndiceHistory
from flask import current_app
from datetime import timedelta

@current_app.cli.command('prune-history')
def prune_history():
    date_ = today() - timedelta(days=7)
    IndiceHistory.query.filter(IndiceHistory.date_ <= date_).delete()
    current_app.logger.info(f'Data before {date_} is removed from history')



from indice_pollution.extensions import db
from importlib import import_module
from indice_pollution.extensions import cache

class Zone(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    id = db.Column(db.Integer, primary_key=True)
    type = db.Column(db.String)
    code = db.Column(db.String)

    @classmethod
    def get(cls, code, type_):
        return Zone.query.filter_by(code=code, type=type_).first()

    @property
    @cache.memoize(timeout=0)
    def lib(self, with_preposition=True, with_article=True):
        types = {
            "region": {"article": "la ", "preposition": "région", "module": "region", "clsname": "Region"},
            "epci": {"article": "l’", "preposition": "EPCI" , "module": "epci", "clsname": "EPCI"},
            "departement": {"article": "le ", "preposition": "département", "module": "departement", "clsname": "Departement"},
            "bassin_dair": {"article": "le ", "preposition": "bassin d’air", "module": "bassin_dair", "clsname": "BassinDAir"},
            "commune": {"article": "la ", "preposition": "commune", "module": "commune", "clsname": "Commune"},
        }
        t = types.get(self.type)
        if not t:
            return ""
        m = import_module(f"indice_pollution.history.models.{t['module']}")
        c = getattr(m, t["clsname"])
        o = db.session.query(c).filter(c.zone_id == self.id).first()
        if not o:
            return ""
        r = ""
        if with_preposition:
            if with_article:
                r = t["article"]
            r += t["preposition"] + " "
        if hasattr(o, "preposition"):
            r += o.preposition + " "
        r += o.nom
        return r





import re
from bs4 import BeautifulSoup
import requests
from . import EpisodeMixin, ForecastMixin

class Service(object):
    is_active = True
    website = 'http://www.ligair.fr/'
    nom_aasqa = "Lig'Air"
    licence = 'OdbL v1.0'


class Forecast(Service, ForecastMixin):
    url = 'https://services1.arcgis.com/HzzPcgRsxxyIZdlU/arcgis/rest/services/Indices_%C3%A0_la_commune_sur_3_jours/FeatureServer/0/query'
    outfields = ['date_ech', 'code_qual', 'lib_qual' , 'coul_qual', 'date_dif', 'code_zone', 'lib_zone']

    def params(self, date_, insee):
        return {
            'where': f"(date_ech >= CURRENT_DATE - INTERVAL '2' DAY) AND code_zone ={insee}",
            'outFields': ",".join(self.outfields),
            'f': 'json',
            'outSR': '4326'
        }

    def getter(self, attributes):
        return super().getter(
            {
                'couleur': attributes.get('coul_qual'),
                **attributes
            }
        )

    def get_from_scraping(self, previous_results, date_, insee):
        r = requests.get(f'http://www.ligair.fr/commune/{self.get_nom_ville(insee)}')
        soup = BeautifulSoup(r.text, "html.parser")
        indices = soup.find_all('div', class_='atmo-index-legend')
        legend = indices[0]
        bars = legend.find_all('div', class_='atmo-bar')
        today = next(filter(lambda v: "Aujourd'hui" in v.text, bars))
        labels = today.find_all('div', class_="atmo-bar-label")
        today_text = next(filter(lambda v: "Aujourd'hui" in v.find_next('strong').text, labels)).text
        indice = today_text[len("Aujourd'hui : "):].lower()
        return previous_results + [
            self.getter({
                "date": str(date_), 
                "indice": indice,
            })
        ]

    def get_nom_ville(self, insee):
        r = requests.get(f'https://geo.api.gouv.fr/communes/{insee}',
                params={
                    "fields": "codesPostaux",
                    "format": "json",
                    "geometry": "centre"
                }
        )
        code_postal = r.json()['codesPostaux'][0]
        r = requests.get('http://www.ligair.fr/ville/city',
                params={"q": code_postal},
                headers={"X-Requested-With": "XMLHttpRequest"}
        )
        return list(r.json().values())[0]

    params_fetch_all = {
        'where': "1=1",
        'outFields': "*",
        'f': 'json',
        'outSR': '4326'
    }

class Episode(Service, EpisodeMixin):
    url = "https://services1.arcgis.com/HzzPcgRsxxyIZdlU/arcgis/rest/services/alerte_3j1/FeatureServer/0/query"


from indice_pollution.extensions import db

class IndiceHistory(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    id = db.Column(db.Integer, primary_key=True)
    date_ = db.Column(db.Date, nullable=False)
    insee = db.Column(db.String, nullable=False)
    features = db.Column(db.JSON, nullable=False)

    @classmethod
    def get(cls, date_, insee):
        return db.session.query(cls).filter_by(date_=date_, insee=insee).first()
    
    @classmethod
    def get_bulk(cls, date_, insee_list):
        return cls.query\
            .filter(cls.date_==date_, cls.insee.in_(set(insee_list)))\
            .all()

    @classmethod
    def get_after(cls, date_, insee):
        return [v.features for v in cls.query.filter(cls.date_ >= date_, cls.insee==insee).all()]

    @classmethod
    def get_or_create(cls, date_, insee, features):
        result =  cls.get(date_, insee)
        if result:
            return result
        result = cls(date_=date_, insee=insee, features=features)
        db.session.add(result)
        return result


from sqlalchemy.orm import relationship
from sqlalchemy.sql import and_
from sqlalchemy.sql.functions import coalesce
from sqlalchemy import func
from indice_pollution.history.models.zone import Zone
from requests.models import codes
from indice_pollution.extensions import db
from indice_pollution.helpers import today
from indice_pollution.history.models import Commune, EPCI
from sqlalchemy import Date, text
from dataclasses import dataclass
from datetime import datetime
from operator import attrgetter, itemgetter


@dataclass
class IndiceATMO(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    zone_id: int = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'), primary_key=True, nullable=False)
    zone: Zone = relationship("indice_pollution.history.models.zone.Zone")
    date_ech: datetime = db.Column(db.DateTime, primary_key=True, nullable=False)
    date_dif: datetime = db.Column(db.DateTime, primary_key=True, nullable=False)
    no2: int = db.Column(db.Integer)
    so2: int = db.Column(db.Integer)
    o3:int = db.Column(db.Integer)
    pm10: int = db.Column(db.Integer)
    pm25: int = db.Column(db.Integer)
    valeur: int = db.Column(db.Integer)

    @classmethod
    def get(cls, insee=None, code_epci=None, date_=None):
        zone_subquery = cls.zone_subquery(insee=insee, code_epci=code_epci).limit(1)
        zone_subquery_or = cls.zone_subquery_or(insee=insee).limit(1)
        zones_subquery = zone_subquery.union(zone_subquery_or)
        date_ = date_ or today()
        query = IndiceATMO\
            .query.filter(
                IndiceATMO.date_ech.cast(Date)==date_,
               IndiceATMO.zone_id.in_(zones_subquery)
            )\
            .order_by(IndiceATMO.date_dif.desc())
        return query.first()

    @classmethod
    def bulk_query(cls, insees=None, date_=None):
        date_ = date_ or today()
        return text(
            """
            SELECT 
                DISTINCT ON (i.date_ech, coalesce(c.insee, c2.insee)) i.*, i.date_ech date, coalesce(c.insee, c2.insee) insee
            FROM
                indice_schema."indiceATMO" i
            JOIN indice_schema.zone z ON i.zone_id = z.id
            LEFT JOIN indice_schema.commune c ON z.type = 'commune' AND z.id = c.zone_id
            LEFT JOIN indice_schema.epci e ON z.type = 'epci' AND z.id = e.zone_id
            LEFT JOIN indice_schema.commune c2 ON c2.epci_id = e.id
            WHERE date(date_ech) = :date_ech AND ((c.insee = ANY(:insees)) OR (c2.insee = ANY(:insees)))
            ORDER BY i.date_ech, coalesce(c.insee, c2.insee), i.date_dif DESC;
            """
        ).bindparams(
            date_ech=date_,
            insees=insees
        )

    @classmethod
    def bulk(cls, insees=None, codes_epci=None, date_=None):
        return db.session.execute(IndiceATMO.bulk_query(insees=insees, date_=date_))

    @classmethod
    def get_all_query(cls, date_):
        commune_alias = db.aliased(Commune)
        commune_id = coalesce(Commune.id, commune_alias.id)
        return db.session.query(
                commune_id, IndiceATMO
            ).select_from(
                IndiceATMO
            ).join(
                Zone
            ).join(
                Commune, and_(Zone.type == 'commune', Zone.id == Commune.zone_id), isouter=True
            ).join(
                EPCI, and_(Zone.type == 'epci', Zone.id == EPCI.zone_id), isouter=True
            ).join(
                commune_alias, commune_alias.epci_id == EPCI.id, isouter=True
            ).filter(
                func.date(IndiceATMO.date_ech) == date_
            ).order_by(
                IndiceATMO.date_ech
            ).distinct(
                IndiceATMO.date_ech, commune_id
            )

    @classmethod
    def get_all(cls, date_):
        return dict(cls.get_all_query(date_).all())

    @classmethod
    def zone_subquery(cls, insee=None, code_epci=None):
        if insee:
            return Commune.get_query(insee=insee).with_entities(Commune.zone_id)
        elif code_epci:
            return EPCI.get_query(code=code_epci).with_entities(EPCI.zone_id)

    @classmethod
    def zone_subquery_or(cls, insee=None, code_epci=None):
        if insee:
            return EPCI.get_query(insee=insee).with_entities(EPCI.zone_id)
        elif code_epci:
            return Commune.get_query(code=code_epci).with_entities(Commune.zone_id)

    @classmethod
    def couleur_from_valeur(cls, valeur):
        return {
            "bon": "#50F0E6",
            "moyen": "#50CCAA",
            "degrade" :"#F0E641",
            "mauvais": "#FF5050",
            "tres_mauvais": "#960032",
            "extrement_mauvais": "#960032",
        }.get(cls.indice_from_valeur(valeur))
    @property
    def couleur(self):
        return self.couleur_from_valeur(self.valeur)

    @classmethod
    def label_from_valeur(cls, valeur):
        return {
            "bon": "Bon",
            "moyen": "Moyen",
            "degrade": "Dégradé",
            "mauvais": "Mauvais",
            "tres_mauvais": "Très mauvais",
            "extrement_mauvais": "Extrêment mauvais",
        }.get(cls.indice_from_valeur(valeur))
    @property
    def label(self):
        return self.label_from_valeur(self.valeur)

    @classmethod
    def indice_from_valeur(cls, valeur):
        return [
            "bon",
            "moyen",
            "degrade",
            "mauvais",
            "tres_mauvais",
            "extrement_mauvais",
        ][valeur - 1]
    @property
    def indice(self):
        return self.indice_from_valeur(self.valeur)

    @classmethod
    def indice_dict(cls, valeur):
        return {
            'indice': cls.indice_from_valeur(valeur),
            'label': cls.label_from_valeur(valeur),
            'couleur': cls.couleur_from_valeur(valeur),
            'valeur': valeur
        }

    @classmethod
    def make_sous_indice_dict(cls, code, valeur):
        if valeur is None:
            return {}
        return {
            **{'polluant_name': code.upper()},
            **cls.indice_dict(valeur)
        }
    @property
    def sous_indices(self):
        return [self.make_sous_indice_dict(k, getattr(self, k)) for k in ['pm25', 'pm10', 'no2', 'o3', 'so2']]

    def dict(self):
        return self.make_dict(self)

    @classmethod
    def make_dict(cls, indice):
        getter = itemgetter if type(indice) == dict else attrgetter
        return {
            **{
                'sous_indices': [cls.make_sous_indice_dict(k, getter(k)(indice)) for k in ['pm25', 'pm10', 'no2', 'o3', 'so2']],
                'date': getter('date_ech')(indice).date().isoformat(),
                'valeur': getter('valeur')(indice)
            },
            **cls.indice_dict(getter('valeur')(indice))
        }

    @property
    def insee(self):
        zone = Zone.query.get(self.zone_id)
        if zone.type == 'commune':
            return zone.code
        if zone.type == 'epci':
            r = db.session.query(
                EPCI, Commune
            ).filter(
                EPCI.zone_id == self.zone_id
            ).limit(1).first()
            return r.Commune.insee

    @property
    def details(self):
        return [self.make_sous_indice_dict(k, getattr(self, k)) for k in ['no2', 'so2', 'o3', 'pm10', 'pm25']]

    @property
    def indice(self):
        return self.indice_from_valeur(self.valeur)


from indice_pollution.extensions import db
from indice_pollution.history.models.commune import Commune
from dataclasses import dataclass

@dataclass
class PotentielRadon(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    zone_id: int = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'), primary_key=True)
    classe_potentiel: int = db.Column(db.Integer)

    @classmethod
    def get(cls, insee):
        return cls.query.join(Commune, Commune.zone_id == cls.zone_id).filter(Commune.insee == insee).first()

    @property
    def label(self):
        return f"Catégorie {self.classe_potentiel}"


from datetime import datetime
import pytz

def today():
    zone = pytz.timezone('Europe/Paris')
    return datetime.now(tz=zone).date()

def oxford_comma(items):
    if not items:
        return
    items = list(items)
    length = len(items)
    if length == 0:
        return
    elif length == 1:
        return items[0]
    elif length == 2:
        return '{} et {}'.format(*items)
    else:
        return '{}, et {}'.format(', '.join(items[:-1]), items[-1])


from dataclasses import dataclass
from indice_pollution.extensions import db
from psycopg2.extras import DateRange
from sqlalchemy.dialects.postgresql import DATERANGE
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy import func
from datetime import date, datetime, timedelta
import os, requests, logging, csv, copy
from indice_pollution.history.models.commune import Commune

from indice_pollution.history.models.departement import Departement

@dataclass
class RAEP(db.Model):
    __tablename__ = "raep"

    id: int = db.Column(db.Integer, primary_key=True)
    zone_id: int = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'), nullable=False)
    validity = db.Column(DATERANGE(), nullable=False)

    cypres: int = db.Column(db.Integer)
    noisetier: int = db.Column(db.Integer)
    aulne: int = db.Column(db.Integer)
    peuplier: int = db.Column(db.Integer)
    saule: int = db.Column(db.Integer)
    frene: int = db.Column(db.Integer)
    charme: int = db.Column(db.Integer)
    bouleau: int = db.Column(db.Integer)
    platane: int = db.Column(db.Integer)
    chene: int = db.Column(db.Integer)
    olivier: int = db.Column(db.Integer)
    tilleul: int = db.Column(db.Integer)
    chataignier: int = db.Column(db.Integer)
    rumex: int = db.Column(db.Integer)
    graminees: int = db.Column(db.Integer)
    plantain: int = db.Column(db.Integer)
    urticacees: int = db.Column(db.Integer)
    armoises: int = db.Column(db.Integer)
    ambroisies: int = db.Column(db.Integer)
    total: int = db.Column(db.Integer)

    __table_args__ = (
        db.Index('raep_zone_validity_idx', zone_id, validity),
        db.UniqueConstraint(zone_id, validity),
        {"schema": "indice_schema"},
    )
    liste_allergenes = ["cypres", "noisetier", "aulne", "peuplier", "saule", "frene", "charme", "bouleau", "platane", "chene", "olivier", "tilleul", "chataignier", "rumex", "graminees", "plantain", "urticacees", "armoises", "ambroisies"]

    @classmethod
    def save_all(cls):
        if not os.getenv('ALLERGIES_URL'):
            return {}
        try:
            r = requests.get(os.getenv("ALLERGIES_URL"))
        except Exception as e:
            logging.error(e)
            return {}
        decoded_content = r.content.decode('utf-8')
        first_column_name = decoded_content[:10] #Il s'agit de la date
        date_format = "%d/%m/%Y"
        debut_validite = datetime.strptime(first_column_name, date_format).date()
        fin_validite = debut_validite + timedelta(weeks=1)
        reader = csv.DictReader(
            decoded_content.splitlines(),
            delimiter=';'
        )

        risques = []
        for r in reader:
            departement_code = f"{r[first_column_name]:0>2}"
            if departement_code == "20":
                departement_code = "2A"   
            departement = Departement.get(departement_code)
            risques.append({
                **{
                    "zone_id": departement.zone_id,
                    "total": r["Total"],
                    "validity": DateRange(debut_validite, fin_validite)
                },
                **{allergene: int(r[allergene]) for allergene in cls.liste_allergenes}
            })
            if departement_code == "2A":
                departement = Departement.get("2B")
                r = copy.deepcopy(risques[-1])
                r["zone_id"] = Departement.get("2B").zone_id
                risques.append(r)
        ins = pg_insert(cls)\
            .values(risques)\
            .on_conflict_do_nothing()
        db.session.execute(ins)
        db.session.commit()

    @classmethod
    def get(cls, insee=None, zone_id=None, date_=None):
        date_ = date_ or date.today()
        if insee:
            if len(insee) == 5:
                zone_id = Commune.get(insee).departement.zone_id
            elif len(insee) == 2:
                zone_id = Departement.get(insee).zone_id
            else:
                return None
        elif zone_id is None:
            return None
        return cls.query.filter(
            cls.zone_id == zone_id,
            cls.validity.contains(date_)
        ).order_by(
            func.upper(cls.validity).desc()
        ).first()

    @classmethod
    def get_all(cls):
        return cls.query.filter(
            cls.validity.contains(date.today())
        ).distinct(cls.zone_id
        ).order_by(
            cls.zone_id,
            func.upper(cls.validity).desc()
        ).all()

    def to_dict(self):
        date_format = "%d/%m/%Y"
        return {
            "total": self.total,
            "allergenes": {
                allergene: getattr(self, allergene)
                for allergene in self.liste_allergenes
            },
            "periode_validite": {
                "debut": self.validity.lower.strftime(date_format),
                "fin": self.validity.upper.strftime(date_format)
            }
        }


from . import ServiceMixin, ForecastMixin, EpisodeMixin

class Service(ServiceMixin):
    is_active = True
    website =  'https://www.qualitaircorse.org/'
    nom_aasqa = 'Qualitair Corse'

class Forecast(Service, ForecastMixin):
    url = 'https://services9.arcgis.com/VQopoXNvUqHYZHjY/arcgis/rest/services/indice_atmo_communal_corse/FeatureServer/0/query'

    params_fetch_all =  {
        "outFields": "*",
        "outSR": "4326",
        "f":"json",
        'orderByFields': 'date_ech DESC',
        'where': "(date_ech >= CURRENT_DATE - INTERVAL '1' DAY)",
    }

class Episode(Service, EpisodeMixin):
    url = 'https://services9.arcgis.com/VQopoXNvUqHYZHjY/arcgis/rest/services/Alrt3j_corse/FeatureServer/0/query'

    params_fetch_all =  {
        "outFields": "*",
        "outSR": "4326",
        "f":"json",
        'orderByFields': 'date_ech DESC',
        'where': "(date_ech >= CURRENT_DATE - INTERVAL '5' DAY)",
    }


from typing import List
from sqlalchemy.orm import relationship
from sqlalchemy import func, select
from indice_pollution.history.models import commune
from indice_pollution.history.models.commune import Commune
from indice_pollution.extensions import db
from datetime import datetime
from indice_pollution.helpers import today
from sqlalchemy import  Date, text

from indice_pollution.history.models.epci import EPCI

class EpisodePollution(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    zone_id: int = db.Column(db.Integer, db.ForeignKey('indice_schema.zone.id'), primary_key=True, nullable=False)
    zone = relationship("indice_pollution.history.models.zone.Zone")
    date_ech: datetime = db.Column(db.DateTime, primary_key=True, nullable=False)
    date_dif: datetime = db.Column(db.DateTime, primary_key=True, nullable=False)
    code_pol: int = db.Column(db.Integer, primary_key=True, nullable=False)
    etat: str = db.Column(db.String)
    com_court: str = db.Column(db.String)
    com_long: str = db.Column(db.String)

    @classmethod
    def get_query(cls, insee=None, code_epci=None, date_=None):
        table_ = cls.__table__
        commune_table = Commune.__table__
        date_ = date_ or today()
        statement = select(table_)
        if insee:
            statement = statement.join(
                commune_table,
                table_.c.zone_id == commune_table.c.zone_pollution_id
            ).filter(
                commune_table.c.insee == insee
            )
        return statement.filter(
            table_.c.date_ech.cast(Date) == date_
        ).order_by(table_.c.date_dif.desc()
        ).fetch(1, with_ties=True)

    @classmethod
    def get(cls, insee=None, code_epci=None, date_=None):
        orms_obj = select(cls).from_statement(cls.get_query(insee, code_epci, date_))
        return list(db.session.execute(orms_obj).scalars())

    @classmethod
    def bulk_query(cls, insees=None, codes_epci=None, date_=None):
        return text(
            """
            SELECT
                DISTINCT ON (e.zone_id, e.date_ech, e.col_pol) e.*, e.date_ech e, c.insee
            FROM
                indice_schema.episode_pollution e
                LEFT JOIN indice_schema.commune c ON c.zone_pollution_id = e.zone_id
                WHERE c.insee = ANY(:insees) AND date_ech=:date_ech
                ORDER BY e.zone_id, e.date_ech, e.date_dif DESC
            """
        ).bindparams(
            date_ech=date_,
            insees=insees
        )
    
    @classmethod
    def bulk(cls, insees=None, codes_epcis=None, date_=None):
        return db.session.execute(cls.bulk_query(insees=insees, date_=date_))

    @classmethod
    def get_all_query(cls, date_):
        return db.session.query(
                cls.zone_id, cls
            ).filter(
                func.date(cls.date_ech) == date_
            ).order_by(
                cls.zone_id, cls.code_pol, cls.date_ech, cls.date_dif
            ).distinct(cls.zone_id, cls.date_ech, cls.code_pol)

    @classmethod
    def get_all(cls, date_):
        to_return = dict()
        for zone_id, e in cls.get_all_query(date_).all():
            to_return.setdefault(zone_id, [])
            to_return[zone_id].append(e)
        return to_return


    @property
    def lib_pol_abbr(self):
        return self.get_lib_pol(self.code_pol)

    @classmethod
    def get_lib_pol_abbr(cls, code_pol):
        return {
            1: 'SO2',
            3: 'NO2',
            4: 'CO',
            5: 'PM10',
            7: 'O3',
            8: 'NO2',
            24: 'PM10',
            39: 'PM25',
        }.get(code_pol)

    @property
    def lib_pol(self):
        return self.get_lib_pol(self.code_pol)

    @classmethod
    def get_lib_pol(cls, code_pol):
        return {
            1: 'Dioxyde de soufre',
            3: 'Dioxyde d’azote',
            4: 'Monoxyde de carbone',
            5: 'Particules PM10',
            7: 'Ozone',
            8: 'Dioxyde d’azote',
            24: 'Particules PM10',
            39: 'Particules PM25',
        }.get(code_pol)

    @property
    def lib_pol_normalized(self):
        return self.get_lib_pol_normalized(self.code_pol)

    @classmethod
    def get_lib_pol_normalized(cls, code_pol):
        return {
            1: 'dioxyde_soufre',
            3: 'dioxyde_azote',
            5: 'particules_fines',
            7: 'ozone',
            8: 'dioxyde_azote',
            24: 'pm10',
            39: 'pm25',
        }.get(code_pol)

    @property
    def code_etat(self):
        return self.get_code_etat(self.etat)

    @classmethod
    def get_code_etat(cls, etat):
        if not isinstance(etat, str):
            return None
        etat = etat.lower()
        if "ssement" in etat:
            return 0
        elif "information" in etat:
            return 1
        elif "alerte" in etat:
            return 2
        else:
            return None

    @property
    def lib_etat(self):
        return self.get_lib_etat(self.code_etat)

    @classmethod
    def get_lib_etat(cls, code_etat):
        return {0: "pas de dépassement", 1: "information", 2: "alerte"}.get(code_etat, "")

    def dict(self):
        return {
            "code_pol": f"{self.code_pol:02}",
            "code_zone": self.zone.code if self.zone else "",
            "com_court": self.com_court,
            "com_long": self.com_long,
            "date": self.date_ech.date().isoformat(),
            "date_ech": self.date_ech.date().isoformat(),
            "date_dif": self.date_dif.date().isoformat(),
            "etat": self.etat,
            "lib_pol": self.lib_pol,
            "lib_pol_normalized": self.lib_pol_normalized
        }

    @classmethod
    def filter_etat_haut(cls, episodes):
        valid_etats = [e.code_etat for e in episodes if e.code_etat != None]
        if len(valid_etats) == 0:
            return []
        max_etat = max(valid_etats)
        if max_etat == 0:
            return []
        return [e for e in episodes if e.code_etat == max_etat]



import requests

def autocomplete(query_string):
    headers = {
        'Accept': 'application/json'
    }
    params = {
        'nom': query_string,
        'fields': 'nom,code,codesPostaux',
        'format': 'json',
        'boost': 'population'
    }
    r = requests.get(
        "https://geo.api.gouv.fr/communes",
        params=params,
        headers=headers
    )

    r.raise_for_status()

    return r.json()


from indice_pollution.extensions import db, cache
from indice_pollution.history.models.departement import Departement
from indice_pollution.history.models.tncc import TNCC
from sqlalchemy.dialects import postgresql
from sqlalchemy.orm import joinedload, relationship
import requests
import json
from flask import current_app
import unicodedata, re

class Commune(db.Model, TNCC):
    __table_args__ = {"schema": "indice_schema"}

    id = db.Column(db.Integer, primary_key=True)
    insee = db.Column(db.String)
    nom = db.Column(db.String)
    epci_id = db.Column(db.Integer, db.ForeignKey("indice_schema.epci.id"))
    epci = relationship("indice_pollution.history.models.epci.EPCI")
    departement_id = db.Column(db.Integer, db.ForeignKey("indice_schema.departement.id"))
    departement = relationship("indice_pollution.history.models.departement.Departement")
    code_zone = db.Column(db.String)
    zone_id = db.Column(db.Integer, db.ForeignKey("indice_schema.zone.id"))
    zone = relationship("indice_pollution.history.models.zone.Zone", foreign_keys=zone_id)
    _centre = db.Column('centre', db.String)
    zone_pollution_id = db.Column(db.Integer, db.ForeignKey("indice_schema.zone.id"))
    zone_pollution = relationship("indice_pollution.history.models.zone.Zone", foreign_keys=zone_pollution_id)
    pollinarium_sentinelle = db.Column(db.Boolean)
    codes_postaux = db.Column(postgresql.ARRAY(db.String))

    def __init__(self, **kwargs):
        if 'code' in kwargs:
            kwargs['insee'] = kwargs.pop('code')
        if 'codeDepartement' in kwargs:
            kwargs['departement'] = Departement.get(kwargs.pop('codeDepartement'))
        super().__init__(**kwargs)

    @property
    def centre(self):
        return json.loads(self._centre)

    @centre.setter
    def centre(self, value):
        self._centre = json.dumps(value)

    @classmethod
    @cache.memoize(timeout=0)
    def get_from_id(cls, id):
        return db.session.query(
            cls
        ).options(
            joinedload(cls.departement).joinedload(Departement.region)
        ).populate_existing(
        ).get(
            id
        )

    @classmethod
    def get(cls, insee):
        if insee is None:
            return None
        return cls.get_query(insee).first()

    @classmethod
    def get_query(cls, insee=None, code=None):
        if insee:
            return db.session.query(cls).filter_by(insee=insee).limit(1)
        elif code:
            from indice_pollution.history.models.epci import EPCI
            subquery = EPCI.get_query(code=code).with_entities(EPCI.id).limit(1).subquery()
            return cls.query.filter(cls.epci_id==subquery).limit(1)

    @classmethod
    def bulk_query(cls, insees=None, codes=None):
        if insees:
            return db.session.query(cls).filter(cls.insee.in_(insees))
        elif codes:
            from indice_pollution.history.models.epci import EPCI
            subquery = EPCI.bulk_query(codes=codes).with_entities(EPCI.id).subquery()
            return cls.query.filter(cls.epci_id.in_(subquery)).limit(1)

    @classmethod
    def get_and_init_from_api(cls, insee):
        res_api = cls.get_from_api(insee)
        if not res_api:
            return None
        o = cls(**res_api)
        db.session.add(o)
        db.session.commit()
        return o

    @classmethod
    def get_from_api(cls, insee):
        r = requests.get(
            f'https://geo.api.gouv.fr/communes/{insee}',
            params={
                "fields": "code,nom,codeDepartement,centre,codesPostaux",
                "format": "json",
            }
        )
        try:
            r.raise_for_status()
        except requests.HTTPError as e:
            current_app.logger.error(f"HTTP Error getting commune: '{insee}' {e}")
            return None
        if not 'codeDepartement' in r.json() or not 'centre' in r.json():
            current_app.logger.error(f'Error getting info about: "{insee}" we need "codeDepartement" and "centre" in "{r.json()}"')
            return None
        j = r.json()
        if 'codesPostaux' in j:
            j['codes_postaux'] = j['codesPostaux']
            del j['codesPostaux']
        return j

    @property
    def code(self):
        return self.insee

    @property
    def departement_nom(self):
        if self.departement:
            return self.departement.nom
        return ""

    @property
    def slug(self):
        # lower, replace whitespace and apostrophe by hyphen, replace œ by oe
        slug = self.nom.lower()\
            .replace(' ', '-')\
            .replace('\'', '-').replace(u'\u2019', '-')\
            .replace(u'\u0153', 'oe')
        # remove diacritics
        slug = re.sub(r'[\u0300-\u036f]', '', unicodedata.normalize('NFD', slug))
        return slug


from sqlalchemy.orm import relationship
from indice_pollution.extensions import db

class BassinDAir(db.Model):
    __table_args__ = {"schema": "indice_schema"}

    id = db.Column(db.Integer, primary_key=True)
    nom = db.Column(db.String)
    code = db.Column(db.String)
    zone_id = db.Column(db.Integer, db.ForeignKey("indice_schema.zone.id"))
    zone = relationship("indice_pollution.history.models.zone.Zone", foreign_keys=zone_id)